{*************************************************************************}
{* 4-bit full-adder (4-bit incrementer) unit for LogicGates8 application *}
{*************************************************************************}

unit adder4f;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, ExtCtrls, StdCtrls, Grids, common;

type
  {***********}
  { TfAdder4f }
  {***********}
  TfAdder4f = class(TForm)
    Shape1: TShape;
    StaticText1: TStaticText;
    Label1, Label2, Label8, Label9, Label10, Label18, Label3, Label11: TLabel;
    Shape20, Shape21, Shape22, Shape23, Shape25, Shape28, Shape32: TShape;
    Shape36, Shape38, Shape45, Shape50, Shape51: TShape;
    Shape52, Shape53, Shape54, Shape55, Shape57, Shape58, Shape60, Shape61: TShape;
    Shape62, Shape64, Shape65, Shape66, Shape67, Shape68, Shape69: TShape;
    Shape70, Shape71, Shape72, Shape73, Shape74, Shape75, Shape76, Shape77: TShape;
    Shape78, Shape79, Shape80, Shape81, Shape82, Shape83, Shape85: TShape;
    Shape86, Shape87, Shape88, Shape89, Shape90, Shape91, Shape92, Shape93: TShape;
    Shape94, Shape95, Shape96, Shape97, Shape98, Shape99, Shape100, Shape101: TShape;
    Shape102, Shape103, Shape104, Shape106, Shape108: TShape;
    Shape110, Shape112, Shape114, Shape116, Shape117: TShape;
    Shape118, Shape119, Shape121, Shape123: TShape;
    Shape126, Shape127, Shape128, Shape56: TShape;
    Image2, Image3, Image4, Image5: TImage;
    laA0, laA1, laA2, laA3, laB0, laB1: TLabel;
    laB2, laB3, laS0, laS1, laS2, laS3: TLabel;
    imSwitchA0, imSwitchA1, imSwitchA2, imSwitchA3: TImage;
    shSwitchA0a, shSwitchA0b, shSwitchA1a, shSwitchA1b: TShape;
    shSwitchA2a, shSwitchA2b, shSwitchA3a, shSwitchA3b: TShape;
    shLedA0, shLedA1, shLedA2, shLedA3: TShape;
    shLedB0, shLedB1, shLedB2, shLedB3: TShape;
    shLedS0, shLedS1, shLedS2, shLedS3, shLedC: TShape;
    sgAdd: TStringGrid;
    Image1: TImage;
    btClose: TButton;
    procedure FormCreate(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure imSwitchA0Click(Sender: TObject);
    procedure imSwitchA1Click(Sender: TObject);
    procedure imSwitchA2Click(Sender: TObject);
    procedure imSwitchA3Click(Sender: TObject);
    procedure btCloseClick(Sender: TObject);
  private
    aLeds: TLeds;
    aSwitches: TSwitches;
    aInput, aOutput: TStatus;
  end;

const
  // Switches description: Data = object positions (transparent image, long bar shape, short bar shape)
  DSwitches: array[0..3] of TSwitchData = (
    (Name: 'SwitchA0';
     Data: (512, 534, 514)),
    (Name: 'SwitchA1';
     Data: (378, 400, 380)),
    (Name: 'SwitchA2';
     Data: (246, 268, 248)),
    (Name: 'SwitchA3';
     Data: (112, 134, 114))
  );

var
  fAdder4f: TfAdder4f;

implementation

{$R *.lfm}

{ 4-bit adder: Set switch position; do binary addition; turn LEDs on/off; fill-in stringgrid }

procedure SwitchAndAdd(DSwitches: array of TSwitchData; Switches: TSwitches; var InputStatus: TStatus;
  Leds: TLeds; var OutputStatus: TStatus; IX: Integer);

const
  Bits: array[False..True] of Integer = (0, 1);

var
  Exp, I: Integer;
  A, B, C, Sout, Cout, S1, S2, C1, C2: Boolean;
  Dec: array[0..2] of Integer;

begin
  // Change current input from 0 to Vs (and vice-versa)
  InputStatus[IX] := not InputStatus[IX];
  // Corresponding LED status
  OutputStatus[IX] := InputStatus[IX];
  // Set switch positions
  SwitchesOnOff(DSwitches, Switches, InputStatus);
  // Do the addition, using 4 1-bit full-adders, as is done in a real circuit:
  // The sum output of a given adder will be the sum value for that bit,
  // the carry output being used as input for the subsequent adder
  Cout := False;                                                               // Cin for first adder set to 0V
  for I := 0 to 3 do begin
    A := OutputStatus[I]; B := OutputStatus[I + 4];                            // actual bit for A and B input
    C := Cout;                                                                 // Cin = previous Cout
    // 1-bit-full-adder binary addition
    Adder(A, B, C, S1, C1, S2, C2, Sout, Cout);
    // Actual bit sum LED status
    OutputStatus[I + 8] := SOut;
  end;
  // Last Cout is carry value for 4-bit operation: Cout LED status
  OutputStatus[12] := Cout;
  // Turn all LEDs on/off
  LedsOnOff(Leds, OutputStatus);
  // Fill-in binary values and calculate decimal values
  Dec[0] := 0; Dec[1] := 0; Dec[2] := 0; Exp := 1;
  for I := 0 to 3 do begin
    fAdder4f.sgAdd.Cells[5 - I, 1] := ' ' + IntToStr(Bits[OutputStatus[I]]);
    fAdder4f.sgAdd.Cells[5 - I, 2] := ' ' + IntToStr(Bits[OutputStatus[I + 4]]);
    fAdder4f.sgAdd.Cells[5 - I, 3] := ' ' + IntToStr(Bits[OutputStatus[I + 8]]);
    Dec[0] += Bits[OutputStatus[I]] * Exp; Dec[1] += Bits[OutputStatus[I + 4]] * Exp; Dec[2] += Bits[OutputStatus[I + 8]] * Exp;
    Exp *= 2;
  end;
  // Display the carry only if it is 1
  if OutputStatus[12] then
    fAdder4f.sgAdd.Cells[1, 3] := ' ' + IntToStr(Bits[OutputStatus[12]])
  else
    fAdder4f.sgAdd.Cells[1, 3] := '';
  Dec[2] += Bits[OutputStatus[12]] * 16;
  // Fill-in decimal values
  for I := 0 to 2 do begin
    fAdder4f.sgAdd.Cells[6, I + 1] := '  ' + IntToStr(Dec[I]);
    if Dec[I] < 10 then
      fAdder4f.sgAdd.Cells[6, I + 1] := ' ' + fAdder4f.sgAdd.Cells[6, I + 1];
  end;
end;

{***********}
{ TfAdder4f }
{***********}

{ Application start: Initialisation }

procedure TfAdder4f.FormCreate(Sender: TObject);

begin
  // Arrays with switches (transparent image and bar shapes) and their "on/off" status
  SetLength(aSwitches, 4); SetLength(aInput, 8);
  aSwitches[0].Switch0 := imSwitchA0; aSwitches[0].Switch1 := shSwitchA0a; aSwitches[0].Switch2 := shSwitchA0b;
  aSwitches[1].Switch0 := imSwitchA1; aSwitches[1].Switch1 := shSwitchA1a; aSwitches[1].Switch2 := shSwitchA1b;
  aSwitches[2].Switch0 := imSwitchA2; aSwitches[2].Switch1 := shSwitchA2a; aSwitches[2].Switch2 := shSwitchA2b;
  aSwitches[3].Switch0 := imSwitchA3; aSwitches[3].Switch1 := shSwitchA3a; aSwitches[3].Switch2 := shSwitchA3b;
  // Arrays with LEDs (shape) and their "on/off" status
  SetLength(aLeds, 13); SetLength(aOutput, 13);
  aLeds[0] := shLedA0; aLeds[1] := shLedA1; aLeds[2] := shLedA2; aLeds[3] := shLedA3;
  aLeds[4] := shLedB0; aLeds[5] := shLedB1; aLeds[6] := shLedB2; aLeds[7] := shLedB3;
  aLeds[8] := shLedS0; aLeds[9] := shLedS1; aLeds[10] := shLedS2; aLeds[11] := shLedS3; aLeds[12] := shLedC;
  // Use subscripts in input/output labels
  laA0.Caption := 'A' + SUB_Digits[0]; laA1.Caption := 'A' + SUB_Digits[1];
  laA2.Caption := 'A' + SUB_Digits[2]; laA3.Caption := 'A' + SUB_Digits[3];
  laB0.Caption := 'B' + SUB_Digits[0]; laB1.Caption := 'B' + SUB_Digits[2];
  laB2.Caption := 'B' + SUB_Digits[2]; laB3.Caption := 'B' + SUB_Digits[3];
  laS0.Caption := 'S' + SUB_Digits[0]; laS1.Caption := 'S' + SUB_Digits[1];
  laS2.Caption := 'S' + SUB_Digits[2]; laS3.Caption := 'S' + SUB_Digits[3];
end;

{ Application show-up: Initialize 4-bit incrementer simulation }

procedure TfAdder4f.FormActivate(Sender: TObject);

var
  I: Integer;

begin
  // All inputs for A, the number to be incremented, to logic 0
  // For B, the increment, least significant bit to logic 1 (increment = addition of 0001)
  for I := 0 to Length(aInput) - 1 do begin
    if I = 4 then
      aInput[I] := True
    else
      aInput[I] := False;
  end;
  SwitchesOnOff(DSwitches, aSwitches, aInput);
  // All outputs to logic 0, except B0 (least significant bit of increment) and S0 (0 + 1 = 1)
  for I := 0 to Length(aOutput) - 1 do begin
    if (I = 4) or (I = 8) then
      aOutput[I] := True
    else
      aOutput[I] := False;
  end;
  LedsOnOff(aLeds, aOutput);
end;

{ Switch A0, A1, A2, A3 activated: Do 4-bit incrementer simulation for actual switch positions }

procedure TfAdder4f.imSwitchA0Click(Sender: TObject);

begin
  SwitchAndAdd(DSwitches, aSwitches, aInput, aLeds, aOutput, 0);               // add 4-bit binary A and 0001
end;

procedure TfAdder4f.imSwitchA1Click(Sender: TObject);

begin
  SwitchAndAdd(DSwitches, aSwitches, aInput, aLeds, aOutput, 1);
end;

procedure TfAdder4f.imSwitchA2Click(Sender: TObject);

begin
  SwitchAndAdd(DSwitches, aSwitches, aInput, aLeds, aOutput, 2);
end;

procedure TfAdder4f.imSwitchA3Click(Sender: TObject);

begin
  SwitchAndAdd(DSwitches, aSwitches, aInput, aLeds, aOutput, 3);
end;

{ Button "Close": Close the window }

procedure TfAdder4f.btCloseClick(Sender: TObject);

begin
  Close;
end;

end.

