{**************************************************************************************************************}
{* 4-bit incrementer (simplified circuit by contraction of 4-bit full-adder) unit for LogicGates8 application *}
{**************************************************************************************************************}

unit adder4fct;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls, ExtCtrls, Grids, common;

type
  {*************}
  { TfAdder4fCt }
  {*************}
  TfAdder4fCt = class(TForm)
    Shape1: TShape;
    StaticText1: TStaticText;
    Label1, Label2, Label3, Label8, Label10, Label18: TLabel;
    Shape20, Shape21, Shape22, Shape23, Shape25, Shape28, Shape32, Shape36, Shape38: TShape;
    Shape45, Shape46, Shape47, Shape48, Shape55, Shape57, Shape58, Shape59, Shape60: TShape;
    Shape61, Shape69, Shape70, Shape73, Shape74, Shape75, Shape76, Shape77, Shape78: TShape;
    Shape79, Shape80, Shape81, Shape82, Shape83, Shape84, Shape85, Shape86, Shape87: TShape;
    Shape88, Shape89, Shape90, Shape91, Shape92, Shape93, Shape94, Shape95, Shape96: TShape;
    Shape97, Shape98, Shape99, Shape100, Shape101, Shape102, Shape103, Shape104, Shape105: TShape;
    Shape106, Shape107, Shape108, Shape109, Shape110, Shape111, Shape112, Shape113, Shape114: TShape;
    Shape115, Shape116, Shape117, Shape119, Shape121, Shape123: TShape;
    Image1, Image2, Image3, Image4, Image5, Image6, Image7: TImage;
    laA0, laA1, laA2, laA3, laS0, laS1, laS2, laS3: TLabel;
    imSwitchA0, imSwitchA1, imSwitchA2, imSwitchA3: TImage;
    shSwitchA0a, shSwitchA0b, shSwitchA1a, shSwitchA1b: TShape;
    shSwitchA2a, shSwitchA2b, shSwitchA3a, shSwitchA3b: TShape;
    shLedA0, shLedA1, shLedA2, shLedA3: TShape;
    shLedS0, shLedS1, shLedS2, shLedS3, shLedC: TShape;
    sgAdd: TStringGrid;
    edContraction: TMemo;
    btClose: TButton;
    procedure FormCreate(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure imSwitchA0Click(Sender: TObject);
    procedure imSwitchA1Click(Sender: TObject);
    procedure imSwitchA2Click(Sender: TObject);
    procedure imSwitchA3Click(Sender: TObject);
    procedure btCloseClick(Sender: TObject);
  private
    aLeds: TLeds;
    aSwitches: TSwitches;
    aInput, aOutput: TStatus;
  end;

const
  // Switches description: Data = objects positions (transparent image, long bar shape, short bar shape)
  DSwitches: array[0..3] of TSwitchData = (
    (Name: 'SwitchA0';
     Data: (480, 502, 482)),
    (Name: 'SwitchA1';
     Data: (346, 368, 348)),
    (Name: 'SwitchA2';
     Data: (214, 236, 216)),
    (Name: 'SwitchA3';
     Data: (80, 102, 82))
  );

var
  fAdder4fCt: TfAdder4fCt;

implementation

{$R *.lfm}

{ 4-bit incrementer: Set switch position; do binary increment; turn LEDs on/off; fill-in stringgrid }

procedure SwitchAndIncrement(DSwitches: array of TSwitchData; Switches: TSwitches; var InputStatus: TStatus;
  Leds: TLeds; var OutputStatus: TStatus; IX: Integer);

// Calculation done as does the real circuit: Usage of an Inverter and 3 XOR/AND gates

const
  Bits: array[False..True] of Integer = (0, 1);

var
  Exp, I: Integer;
  Cout: Boolean;
  Dec: array[0..1] of Integer;

begin
  // Change current input from 0 to Vs (and vice-versa)
  InputStatus[IX] := not InputStatus[IX];
  // Corresponding LED status
  OutputStatus[IX] := InputStatus[IX];
  // Set switch positions
  SwitchesOnOff(DSwitches, Switches, InputStatus);
  // Do the increment using Boolean algebra (based on logic gates of the circuit)
  OutputStatus[4] := not InputStatus[0];                                       // S0 = not A0
  Cout := InputStatus[0];                                                      // carry for bit 0 addition = A0
  for I := 1 to 3 do begin
    // Do addition for bits 1-3
    OutputStatus[4 + I] := InputStatus[I] xor Cout;                            // S[i] = A[i] xor C[i] (C[i] = carry of previous addition)
    Cout := InputStatus[I] and Cout;                                           // carry for this bit addition: C[i+1] = A[i] and C[i]
  end;
  // Last Cout is carry value for 4-bit operation: Cout LED status
  OutputStatus[8] := Cout;
  // Turn all LEDs on/off
  LedsOnOff(Leds, OutputStatus);
  // Fill-in binary values and calculate decimal values
  Dec[0] := 0; Dec[1] := 0; Exp := 1;
  for I := 0 to 3 do begin
    fAdder4fCt.sgAdd.Cells[5 - I, 1] := ' ' + IntToStr(Bits[OutputStatus[I]]);
    fAdder4fCt.sgAdd.Cells[5 - I, 2] := ' ' + IntToStr(Bits[OutputStatus[I + 4]]);
    Dec[0] += Bits[OutputStatus[I]] * Exp; Dec[1] += Bits[OutputStatus[I + 4]] * Exp;
    Exp *= 2;
  end;
  // Display the carry only if it is 1
  if OutputStatus[8] then
    fAdder4fCt.sgAdd.Cells[1, 2] := ' ' + IntToStr(Bits[OutputStatus[8]])
  else
    fAdder4fCt.sgAdd.Cells[1, 2] := '';
  Dec[1] += Bits[OutputStatus[8]] * 16;
  // Fill-in decimal values
  for I := 0 to 1 do begin
    fAdder4fCt.sgAdd.Cells[6, I + 1] := '  ' + IntToStr(Dec[I]);
    if Dec[I] < 10 then
      fAdder4fCt.sgAdd.Cells[6, I + 1] := ' ' + fAdder4fCt.sgAdd.Cells[6, I + 1];
  end;
end;

{*************}
{ TfAdder4fCt }
{*************}

procedure TfAdder4fCt.FormCreate(Sender: TObject);

begin
  // Arrays with switches (transparent image and bar shapes) and their "on/off" status
  SetLength(aSwitches, 4); SetLength(aInput, 4);
  aSwitches[0].Switch0 := imSwitchA0; aSwitches[0].Switch1 := shSwitchA0a; aSwitches[0].Switch2 := shSwitchA0b;
  aSwitches[1].Switch0 := imSwitchA1; aSwitches[1].Switch1 := shSwitchA1a; aSwitches[1].Switch2 := shSwitchA1b;
  aSwitches[2].Switch0 := imSwitchA2; aSwitches[2].Switch1 := shSwitchA2a; aSwitches[2].Switch2 := shSwitchA2b;
  aSwitches[3].Switch0 := imSwitchA3; aSwitches[3].Switch1 := shSwitchA3a; aSwitches[3].Switch2 := shSwitchA3b;
  // Arrays with LEDs (shape) and their "on/off" status
  SetLength(aLeds, 9); SetLength(aOutput, 9);
  aLeds[0] := shLedA0; aLeds[1] := shLedA1; aLeds[2] := shLedA2; aLeds[3] := shLedA3;
  aLeds[4] := shLedS0; aLeds[5] := shLedS1; aLeds[6] := shLedS2; aLeds[7] := shLedS3; aLeds[8] := shLedC;
  // Use subscripts in input/output labels
  laA0.Caption := 'A' + SUB_Digits[0]; laA1.Caption := 'A' + SUB_Digits[1];
  laA2.Caption := 'A' + SUB_Digits[2]; laA3.Caption := 'A' + SUB_Digits[3];
  laS0.Caption := 'S' + SUB_Digits[0]; laS1.Caption := 'S' + SUB_Digits[1];
  laS2.Caption := 'S' + SUB_Digits[2]; laS3.Caption := 'S' + SUB_Digits[3];
  // Use subscripts in contraction description
  edContraction.Text := StringReplace(edContraction.Text, 'A0', 'A' + SUB_Digits[0], [rfReplaceAll]);
  edContraction.Text := StringReplace(edContraction.Text, 'B0', 'B' + SUB_Digits[0], [rfReplaceAll]);
  edContraction.Text := StringReplace(edContraction.Text, 'C0', 'C' + SUB_Digits[0], [rfReplaceAll]);
  edContraction.Text := StringReplace(edContraction.Text, 'S0', 'S' + SUB_Digits[0], [rfReplaceAll]);
  edContraction.Text := StringReplace(edContraction.Text, 'Ā0', 'Ā' + SUB_Digits[0], [rfReplaceAll]);
  edContraction.Text := StringReplace(edContraction.Text, 'A1', 'A' + SUB_Digits[1], [rfReplaceAll]);
  edContraction.Text := StringReplace(edContraction.Text, 'B1', 'B' + SUB_Digits[1], [rfReplaceAll]);
  edContraction.Text := StringReplace(edContraction.Text, 'C1', 'C' + SUB_Digits[1], [rfReplaceAll]);
  edContraction.Text := StringReplace(edContraction.Text, 'S1', 'S' + SUB_Digits[1], [rfReplaceAll]);
  edContraction.Text := StringReplace(edContraction.Text, 'Ai', 'A' + SUB_I, [rfReplaceAll]);
  edContraction.Text := StringReplace(edContraction.Text, 'Bi', 'B' + SUB_I, [rfReplaceAll]);
  edContraction.Text := StringReplace(edContraction.Text, 'Ci+1', 'C' + SUB_I + SUB_Plus + SUB_Digits[1], [rfReplaceAll]);
  edContraction.Text := StringReplace(edContraction.Text, 'Ci', 'C' + SUB_I, [rfReplaceAll]);
  edContraction.Text := StringReplace(edContraction.Text, 'Si', 'S' + SUB_I, [rfReplaceAll]);
end;

{ Application show-up: Initialize 4-bit incrementer simulation }

procedure TfAdder4fCt.FormActivate(Sender: TObject);

var
  I: Integer;

begin
  // All inputs for A, the number to be incremented, to logic 0
  for I := 0 to Length(aInput) - 1 do
    aInput[I] := False;
  SwitchesOnOff(DSwitches, aSwitches, aInput);
    // All outputs to logic 0, except S0 (0 + 1 = 1)
  for I := 0 to Length(aOutput) - 1 do begin
    if I = 4 then
      aOutput[I] := True
    else
      aOutput[I] := False;
  end;
  LedsOnOff(aLeds, aOutput);
end;

{ Switch A0, A1, A2, A3 activated: Do 4-bit incrementer simulation for actual switch positions }

procedure TfAdder4fCt.imSwitchA0Click(Sender: TObject);

begin
  SwitchAndIncrement(DSwitches, aSwitches, aInput, aLeds, aOutput, 0);
end;

procedure TfAdder4fCt.imSwitchA1Click(Sender: TObject);

begin
  SwitchAndIncrement(DSwitches, aSwitches, aInput, aLeds, aOutput, 1);
end;

procedure TfAdder4fCt.imSwitchA2Click(Sender: TObject);

begin
  SwitchAndIncrement(DSwitches, aSwitches, aInput, aLeds, aOutput, 2);
end;

procedure TfAdder4fCt.imSwitchA3Click(Sender: TObject);

begin
  SwitchAndIncrement(DSwitches, aSwitches, aInput, aLeds, aOutput, 3);
end;

{ Button "Close": Close the window }

procedure TfAdder4fCt.btCloseClick(Sender: TObject);

begin
  Close;
end;

end.

